# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import enum
import time
import threading
import logging


log = logging.getLogger(__name__)


class ServiceState(enum.Enum):
    STOPPED = 'stopped'
    STARTING = 'starting'
    RUNNING = 'running'
    PAUSING = 'pausing'
    PREPARING = 'preparing'
    PAUSED = 'paused'
    STOPPING = 'stopping'
    EXCEPTION = 'exception'


STOPPED = ServiceState.STOPPED
STARTING = ServiceState.STARTING
RUNNING = ServiceState.RUNNING
PAUSING = ServiceState.PAUSING
PREPARING = ServiceState.PREPARING
PAUSED = ServiceState.PAUSED
STOPPING = ServiceState.STOPPING
EXCEPTION = ServiceState.EXCEPTION


intermediate_states = {
    RUNNING: PAUSED,
    STOPPED: PAUSED,
}


class Service:
    """
    A wrapper around workers, that you can use to control your workers without
    worrying about threading.
    """

    State = globals()['ServiceState']

    def __init__(self, name, worker):
        self.name = name
        self.worker = worker
        self.exception = None
        self.state_listeners = set()
        self.next_transition = None
        self.state_lock = threading.RLock()
        self._target_state = None
        self._next_state = None
        self._state = STOPPED
        self._transition = None
        self.state_timestamp = time.time()
        worker.service = self

    def start(self):
        """
        Makes sure the worker ends up in the ``RUNNING`` state eventually.
        """
        self._transition_to(RUNNING)

    def pause(self):
        """
        Makes sure the worker ends up in the ``PAUSED`` state eventually.
        """
        self._transition_to(PAUSED)

    def prepare(self):
        """
        An alias for :meth:`pause` for ensuring compaitibility with the
        :class:`Worker` API.
        """
        self._transition_to(PAUSED)

    def stop(self):
        """
        Makes sure the worker ends up in the ``STOPPED`` state eventually.
        """
        self._transition_to(STOPPED)

    def register_state_change_listener(self, callback):
        """
        Registers a `callable` that will be invoked whenever the state of the
        worker changes. The *callback* will receive three arguments:

        * this service,
        * the old state and
        * the new (current) state.

        Note, that due to the nature of threading, it is possible that the
        Service is already in another state than the one provided as the third
        argument.
        """
        self.state_listeners.add(callback)

    def unregister_state_change_listener(self, callback):
        """
        Removes a previously registered listener.
        """
        self.state_listeners.discard(callback)

    def _transition_to(self, target_state):
        with self.state_lock:
            if self._state == EXCEPTION:
                log.debug('_transition_to(%s) -> EXCEPTION' % target_state)
                return
            if self._state == target_state:
                self._target_state = None
                self._next_state = None
                log.debug('_transition_to(%s) -> NOP' % target_state)
                return
            if (self._target_state == target_state and self._transition and
                    self._transition[1] == target_state):
                # already transitioning to give state
                self._target_state = None
                self._next_state = None
                log.debug('_transition_to(%s) -> in progress' % target_state)
                return
            transition = (self._state, target_state)
            if transition in self.worker._state_transitions:
                log.debug('_transition_to(%s) -> transition initiated' %
                          target_state)
                self._target_state = None
                self._next_state = None
                if target_state == RUNNING:
                    self.state = STARTING
                elif target_state == STOPPED:
                    self.state = STOPPING
                elif target_state == PAUSED and self.state == STOPPED:
                    self.state = PREPARING
                elif target_state == PAUSED:
                    self.state = PAUSING
                self._target_state = target_state
                funcname = self.worker._state_transitions[transition]
                callback = getattr(self.worker, funcname)
                self._transition = transition
                threading.Thread(target=self._execute_transition,
                                 args=(transition, callback)).start()
                return
            if target_state in intermediate_states:
                log.debug('_transition_to(%s) -> intermediate(%s)' % (
                    target_state, intermediate_states[target_state]))
                self._transition_to(intermediate_states[target_state])
                self._next_state = target_state
            else:
                log.debug('_transition_to(%s) -> queued' % target_state)
                self._next_state = target_state

    def _execute_transition(self, transition, callback):
        log.debug('_execute_transition(%s, %s)' % (str(transition),
                                                   callback.__name__))
        state_timestamp = self.state_timestamp
        try:
            callback()
            with self.state_lock:
                if self._transition == transition:
                    self._transition = None
                if state_timestamp >= self.state_timestamp:
                    self.state = transition[1]
        except Exception as exception:
            self.set_exception(exception)

    def set_exception(self, exception):
        with self.state_lock:
            if self._state == EXCEPTION:
                return
            old_state = self._state
            self._state = EXCEPTION
            self.exception = exception
            self.state_timestamp = time.time()
        self.worker.cleanup(exception)
        self._state_changed(old_state, EXCEPTION)

    def _state_changed(self, old, new):
        log.debug('changed state: %s -> %s' % (old, new))
        if new == EXCEPTION:
            log.exception(self.exception)
        for callback in self.state_listeners:
            callback(self, old, new)
        for callback in self.worker.state_listeners:
            callback(self, old, new)
        with self.state_lock:
            if self._next_state:
                log.debug('  next state: %s' % (self._next_state))
                next_state = self._next_state
                self._next_state = None
                self._transition_to(next_state)
            elif self._target_state and self._state != self._target_state:
                log.debug('  target state: %s' % (self._target_state))
                self._transition_to(self._target_state)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        with self.state_lock:
            if new_state == self._state or self._state == EXCEPTION:
                return
            old_state = self._state
            self._state = new_state
            self.state_timestamp = time.time()
        self._state_changed(old_state, new_state)



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import asyncio
import json
from collections import OrderedDict
import warnings


class ServiceMonitorProtocol(asyncio.Protocol):

    def __init__(self, conf):
        self._conf = conf
        self.server = None
        self.input_buffer = b''

    def connection_made(self, transport):
        self.transport = transport

    def data_received(self, data):
        self.input_buffer = self.input_buffer + data
        self.handle_input()

    def handle_input(self):
        lines = self.input_buffer.split(b'\n')
        self.input_buffer = lines.pop()
        if not self.server:
            return
        for command in lines:
            command = command.strip()
            if command == b'start':
                self._conf.loop.create_task(self.server.controller.start())
            elif command == b'restart':
                self.server.restart()
            elif command == b'pause':
                self._conf.loop.create_task(self.server.controller.pause())
            elif command == b'stop':
                self._conf.loop.create_task(self.server.stop())
            else:
                warnings.warn('Received invalid command: ' + command)

    def set_instance(self, server):
        assert self.server is None
        self.server = server
        self.server.controller.on('state-change', self._state_change)
        self._conf.loop.create_task(self._send_service_states_async())

    def clear_instance(self, reloading):
        assert self.server is not None
        self.server.controller.off('state-change', self._state_change)
        self.server = None
        if not self.transport:
            return
        if reloading:
            self._send(json.dumps('reloading'))
        else:
            self._send(json.dumps('shutting down'))

    def connection_lost(self, exc):
        self.transport = None
        if self.server:
            self.clear_instance(False)
        self._conf._remove_monitor_connection(self)

    def _state_change(self, services):
        if not services:
            return
        if not self.transport:
            return
        services = OrderedDict((k, v.value) for k, v in services.items())
        self._send(json.dumps(services))

    @asyncio.coroutine
    def _send_service_states_async(self):
        if self.server is None:
            return
        services = yield from self.server.controller.service_states()
        self._state_change(services)

    def _send(self, data):
        self.transport.write(data.encode('UTF-8') + b'\n')



from .worker import Worker, transitions
from .socketserver import SocketServerWorker
from .simple import SimpleWorker
from .asyncio import AsyncioWorker
from .watcher import FileWatcherWorker

__all__ = ('Worker', 'transitions', 'SocketServerWorker', 'SimpleWorker',
           'AsyncioWorker', 'FileWatcherWorker')



from .worker import Worker
import watchdog.observers
import watchdog.events
import watchdog.utils
import os
import warnings
import threading


Observer = watchdog.observers.Observer

if Observer.__name__ == 'InotifyObserver':
    import watchdog.observers.inotify

    # The inotify original observer has a small delay for pairing IN_MOVED_FROM
    # and IN_MOVE_TO events. Since we do not care whether something was moved
    # *into* or *out of* our watches, we will override this behaviour to achieve
    # the fastest possible reload times.

    class InotifyBuffer(watchdog.observers.inotify.InotifyBuffer):

        def __init__(self, *args, **kwargs):
            self.delay = 0
            super().__init__(*args, **kwargs)

        def close(self):
            self.stop()

    class InotifyEmitter(watchdog.observers.inotify.InotifyEmitter):

        def on_thread_start(self):
            path = watchdog.utils.unicode_paths.encode(self.watch.path)
            self._inotify = InotifyBuffer(path, self.watch.is_recursive)

    class InotifyObserver(watchdog.observers.inotify.InotifyObserver):

        def __init__(self, *args, **kwargs):
            watchdog.observers.api.BaseObserver.__init__(
                self, *args, emitter_class=InotifyEmitter, **kwargs)

    Observer = InotifyObserver


class FileWatcherWorker(Worker, watchdog.events.FileSystemEventHandler):
    """
    The implementation of a single service.

    The worker will be wrapped in :class:`Service` objects before being started.
    """

    def watch(self, path):
        if not os.path.exists(path):
            warnings.warn('Cannot watch "%s": path does not exist' % path)
            return
        if os.path.isfile(path):
            dir = os.path.dirname(path)
            self.target_files.append(path)
        else:
            dir = path
            self.target_dirs.append(path)
        if dir in self.observed_dirs:
            return
        with self._observer_lock:
            for other in self.observed_dirs.copy():
                if dir.startswith(other):
                    return
                if other.startswith(dir):
                    if self.observer:
                        self.observer.unschedule(self.observed_dirs[other])
                    del self.observed_dirs[other]
            if self.observer:
                self.observed_dirs[dir] = \
                    self.observer.schedule(self, dir, recursive=True)
            else:
                self.observed_dirs[dir] = None

    def prepare(self):
        self._observer_lock = threading.Lock()
        self.observer = None
        self.observed_dirs = {}
        self.target_files = []
        self.target_dirs = []

    def start(self):
        self.observer = Observer()
        for dir in self.observed_dirs:
            self.observed_dirs[dir] = \
                self.observer.schedule(self, dir, recursive=True)
        self.observer.start()

    def stop(self):
        pass

    def pause(self):
        with self._observer_lock:
            self.observer.stop()
        self.observer.join()
        self.observer = None

    def cleanup(self, exception):
        if hasattr(self, '_observer_lock'):
            with self._observer_lock:
                if hasattr(self, 'observer'):
                    self.observer.stop()
                    self.observer.join()

    def on_any_event(self, event):
        relevant = hasattr(event, 'src_path') and (
            event.src_path in self.target_files or
            any(event.src_path.startswith(dir) for dir in self.target_dirs))
        if relevant:
            self.changed(event.src_path)



import abc
from ..service import Service


STOPPED = Service.State.STOPPED
STARTING = Service.State.STARTING
RUNNING = Service.State.RUNNING
PAUSING = Service.State.PAUSING
PREPARING = Service.State.PREPARING
PAUSED = Service.State.PAUSED
STOPPING = Service.State.STOPPING
EXCEPTION = Service.State.EXCEPTION


final_states = {
    'start': RUNNING,
    'stop': STOPPED,
    'pause': PAUSED,
    'prepare': PAUSED,
}

invalid_transitions = {
    (STARTING, RUNNING),
    (STOPPING, STOPPED),
    (PAUSING, PAUSED),
    (PREPARING, PAUSED),
}


def transitions(state1, state2=None):
    """
    This annotation can add additional transitions to a class.

    If your worker is capable of going from RUNNING to STOPPED, for example, you
    can add the additional transition to a new class method:

    .. code-block:: python

        class MyWorker(Worker):

            @transitions(Service.State.RUNNING, Service.State.STOPPED)
            def kill():
                # ...
    """

    def wrapper(func):
        nonlocal state1, state2
        if not hasattr(func, 'transitions'):
            func.transitions = set()
        name = func.__name__
        if not state2:
            try:
                state2 = final_states[name]
            except KeyError:
                raise Exception(
                    'Function %s() has no end state for transition from %s' %
                    (name, state1))
        elif name in final_states and final_states[name] != state2:
            raise Exception('Function %s() must transition to %s' %
                            (name, final_states[name]))
        for transition in func.transitions:
            if state2 != transition[1]:
                raise Exception(
                    'Function %s() cannot traansition to both %s and %s' %
                    (name, state2, transition[1]))
        func.transitions.add((state1, state2))
        return func
    if state1 == state2 or (state1, state2) in invalid_transitions:
        raise ValueError('Invalid transition: %s' % str((state1, state2)))
    return wrapper


class WorkerMeta(abc.ABCMeta):

    def __init__(cls, name, parents, members):
        transitions = {}
        for name in members:
            func = members[name]
            if not callable(func):
                continue
            if not hasattr(func, 'transitions'):
                continue
            for transition in func.transitions:
                if transition in transitions:
                    raise Exception(
                        'Transition %s already registered as function %s' %
                        (str(transition), transitions[transition]))
                transitions[transition] = name
        if not parents:
            tmp = {
                (STOPPED, PAUSED): 'prepare',
                (PAUSED, RUNNING): 'start',
                (RUNNING, PAUSED): 'pause',
                (PAUSED, STOPPED): 'stop',
            }
            tmp.update(transitions)
            transitions = tmp
        if transitions:
            for parent in parents:
                if hasattr(parent, '_state_transitions'):
                    tmp = parent._state_transitions.copy()
                    tmp.update(transitions)
                    transitions = tmp
                    break
            members['_state_transitions'] = transitions
            cls._state_transitions = transitions
        abc.ABCMeta.__init__(cls, name, parents, members)


class Worker(metaclass=WorkerMeta):
    """
    The implementation of a single service.

    The worker will be wrapped in :class:`Service` objects before being started.
    """

    state_listeners = set()

    @property
    def state(self):
        return self.service.state

    @abc.abstractmethod
    def prepare(self):
        """
        Implements the transition from STOPPED to PAUSED.
        """

    @abc.abstractmethod
    def start(self):
        """
        Implements the transition from PAUSED to RUNNING.
        """

    @abc.abstractmethod
    def stop(self):
        """
        Implements the transition from PAUSED to STOPPED.
        """

    @abc.abstractmethod
    def pause(self):
        """
        Implements the transition from RUNNING to PAUSED.
        """

    @abc.abstractmethod
    def cleanup(self, exception):
        """
        Called when an exception occured. Due to the nature of threading, it is
        not entirely clear, in which state the worker was, when this specific
        exception occurred.
        """

    def register_state_change_listener(self, callback):
        """
        Registers a `callable` that will be invoked whenever the state of this
        worker changes. The *callback* will receive three arguments:

        * a :class:`Service <score.serve.service.Service>` wrapping this worker,
        * the old :class:`state <score.serve.service.State>` and
        * the new (current) state.

        Note, that due to the nature of threading, it is possible that the
        Service is already in another state than the one provided as the third
        argument.
        """
        if not self.state_listeners:
            self.state_listeners = set()
        self.state_listeners.add(callback)

    def unregister_state_change_listener(self, callback):
        """
        Removes a previously registered listener.
        """
        self.state_listeners.discard(callback)



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import socket
import asyncio
from score.init import (
    ConfiguredModule, parse_list, parse_bool, parse_host_port, init_from_file,
    InitializationError)
from .service import Service
from ._forked import fork, Backgrounded
from ._changedetect import ChangeDetector
from collections import OrderedDict
from contextlib import contextmanager
import traceback
import signal
import logging
from .worker import Worker
try:
    import uvloop
except ImportError:
    pass
else:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


log = logging.getLogger('score.serve')

defaults = {
    'autoreload': False,
    'modules': [],
    'monitor': None,
}


def init(confdict):
    """
    Initializes this module acoording to :ref:`our module initialization
    guidelines <module_initialization>` with the following configuration keys:

    :confkey:`autoreload` :confdefault:`False`
        When set to :func:`true <score.init.parse_bool>`, the server will
        automatically reload whenever it detects a change in one of the python
        files, that are in use.

    :confkey:`modules`
        The :func:`list <score.init.parse_list>` of modules to serve. This need
        to be a list of module aliases, i.e. the same name, with which you
        configured the module with ("score.http" becomes "http" if not specified
        otherwise.)

    """
    conf = defaults.copy()
    conf.update(confdict)
    modules = parse_list(conf['modules'])
    if not modules:
        import score.serve
        raise InitializationError(score.serve, 'No modules configured')
    autoreload = parse_bool(conf['autoreload'])
    monitor_host_port = None
    if conf['monitor']:
        monitor_host_port = parse_host_port(conf['monitor'])
    return ConfiguredServeModule(conf['conf'], modules, autoreload,
                                 monitor_host_port)


class ConfiguredServeModule(ConfiguredModule):
    """
    This module's :class:`configuration class`
    """

    def __init__(self, conf, modules, autoreload, monitor_host_port):
        import score.serve
        ConfiguredModule.__init__(self, score.serve)
        self.conf = conf
        self.modules = modules
        self.autoreload = autoreload
        self.monitor_connections = []
        self.monitor_host_port = monitor_host_port
        self.loop = asyncio.new_event_loop()
        self.loop.getaddrinfo = self._getaddrinfo

    def _finalize(self, score):
        self._score = score

    def _getaddrinfo(self, host, port, *, family=0, type=0, proto=0, flags=0):
        """
        The default loop implementation uses loop.run_in_executor to resolve
        addresses, which in turn uses a ThreadPoolExecutor or a
        ProcessPoolExecutor. Neither of these executors is feasable, since
        forking would break both of them. This is the reason we are resolving
        hostnames synchronously.
        """
        @asyncio.coroutine
        def getaddrinfo():
            return socket.getaddrinfo(host, port, family, type, proto, flags)
        return getaddrinfo()

    def start(self):
        """
        Starts all configured workers and runs until the workers stop or
        <CTRL-C> ist pressed. Will optionally reload the server, if it was
        configured to do so via ``autoreload``.
        """
        if self.monitor_host_port:
            coroutine = self.loop.create_server(self._create_monitor_connection,
                                                host=self.monitor_host_port[0],
                                                port=self.monitor_host_port[1])
            self.loop.create_task(coroutine)
        while True:
            self.instance = _ServerInstance(self)
            for connection in self.monitor_connections:
                connection.set_instance(self.instance)
            self.instance.run_until_stopped()
            reload = self.instance.reload
            self.instance = None
            for connection in self.monitor_connections:
                connection.clear_instance(reload)
            if not reload:
                break
            log.info('reloading')

    def _create_monitor_connection(self):
        from .monitor import ServiceMonitorProtocol
        connection = ServiceMonitorProtocol(self)
        if self.instance:
            connection.set_instance(self.instance)
        self.monitor_connections.append(connection)
        return connection

    def _remove_monitor_connection(self, connection):
        self.monitor_connections.remove(connection)

    def _iter_workers(self):
        for descriptor in self.modules:
            if '/' in descriptor:
                module, names = tuple(map(str.strip, descriptor.split('/', 1)))
                names = tuple(map(str.strip, names.split(',')))
            else:
                module = descriptor
                names = None
            response = self._score._modules[module].score_serve_workers()
            if isinstance(response, Worker):
                yield module, response
            elif isinstance(response, list):
                if len(response) == 1:
                    yield module, response[0]
                else:
                    for i, worker in enumerate(response):
                        name = '%s/%d' % (module, i)
                        yield name, worker
            elif isinstance(response, dict):
                for name, worker in response.items():
                    if names is not None and name not in names:
                        continue
                    name = '%s/%s' % (module, name)
                    yield name, worker
            else:
                raise RuntimeError(
                    'Invalid return value of %s.score_serve_workers(): %s' %
                    (module, repr(response)))

    @property
    def workers(self):
        if not hasattr(self, '_workers'):
            self._workers = list(self._iter_workers())
        return self._workers


class _ServerInstance:

    def __init__(self, conf):
        self.conf = conf
        self.loop = conf.loop
        self.controller = fork(self.loop, ServiceController, self.conf)

    def run_until_stopped(self):
        if self.conf.autoreload:
            self.controller.on('restart', self.restart)
        self.reload = None
        self.controller.on('state-change', self.quit_if_stopped)
        # self.loop.set_debug(True)
        self.loop.add_signal_handler(signal.SIGINT, self.signal_handler_stop)
        self.__start_1()
        self.__stopping = False
        self.loop.run_forever()
        self.loop.remove_signal_handler(signal.SIGINT)

    def __start_1(self):
        # make sure that all services are in PAUSED state first.
        # we will start them all at once afterwards.
        task = self.loop.create_task(self.controller.pause())
        task.add_done_callback(self.__start_2)

    def __start_2(self, future):
        exc = future.exception()
        if not exc:
            # start all services at once
            task = self.loop.create_task(self.controller.start())
            task.add_done_callback(lambda *_: log.info('started'))
        elif not self.conf.autoreload:
            self.loop.remove_signal_handler(signal.SIGINT)
            self.loop.create_task(self.stop())

    def signal_handler_stop(self):
        log.info('Ctrl+C detected, stopping')
        self.loop.remove_signal_handler(signal.SIGINT)
        self.loop.create_task(self.stop())
        self.reload = False

    @asyncio.coroutine
    def stop(self):
        if self.__stopping:
            return
        self.__stopping = True
        event = asyncio.Event(loop=self.loop)

        def signal_if_all_stopped(states):
            if not self.all_services_stopped(states):
                return
            self.controller.off('state-change', signal_if_all_stopped)
            task = self.loop.create_task(self.controller.kill())
            task.add_done_callback(wait_on_pending_tasks)

        def wait_on_pending_tasks(future):
            coro = self.wait_on_pending_tasks([current_task])
            task = self.loop.create_task(coro)
            task.add_done_callback(signal_done)

        def signal_done(future):
            event.set()

        current_task = asyncio.Task.current_task(loop=self.loop)
        states = yield from self.controller.service_states()
        if not self.all_services_stopped(states):
            self.controller.off('state-change', self.quit_if_stopped)
            self.controller.on('state-change', signal_if_all_stopped)
            yield from self.controller.stop()
            yield from event.wait()
        self.cleanup()

    def cleanup(self):
        self.controller.cleanup()
        self.loop.stop()

    def all_services_stopped(self, states):
        if isinstance(states, dict):
            states = states.values()
        return all(state in (Service.State.STOPPED, Service.State.EXCEPTION)
                   for state in states)

    def quit_if_stopped(self, states):
        if not self.all_services_stopped(states):
            return
        self.controller.off('state-change', self.quit_if_stopped)
        self.loop.create_task(self.stop())

    def restart(self):
        if self.reload is None:
            self.reload = True
        self.loop.create_task(self.stop())

    @asyncio.coroutine
    def wait_on_pending_tasks(self, ignored_tasks=None):
        event = asyncio.Event(loop=self.loop)
        if ignored_tasks is None:
            ignored_tasks = []
        ignored_tasks.append(asyncio.Task.current_task(loop=self.loop))

        def check_pending(future=None):
            if future:
                # collect the task exception, otherwise the asyncio library will
                # complain
                future.exception()
            pending_tasks = [t for t in asyncio.Task.all_tasks(self.loop)
                             if not t.done()]
            if len(pending_tasks) <= len(ignored_tasks):
                event.set()
                return
            task = pending_tasks.pop()
            while task in ignored_tasks:
                task = pending_tasks.pop()
            task.add_done_callback(check_pending)

        check_pending()
        yield from event.wait()


class ServiceController(Backgrounded):

    def __init__(self, conf):
        self.conf = conf
        self._services = None
        self._changedetector = None

    def start(self):
        if not self._services:
            self._init_services()
        self._call_on_subservices('start')

    def pause(self):
        if not self._services:
            self._init_services()
        self._call_on_subservices('pause')

    def stop(self):
        if not self._services:
            return
        if self._changedetector:
            self._changedetector.stop(wait=False)
            self._changedetector = None
        self._call_on_subservices('stop')

    def service_states(self):
        if not self._services:
            return []
        result = OrderedDict()
        for name, service in self._services.items():
            result[name] = service.state
        return result

    @property
    @contextmanager
    def _acquire_service_locks(self):
        acquired = []
        try:
            for service in self._services.values():
                service.state_lock.acquire()
                acquired.append(service)
            yield
        finally:
            for service in acquired:
                service.state_lock.release()

    def _init_services(self):
        if self.conf.autoreload:
            self._changedetector = ChangeDetector()
            self._changedetector.observe(self.conf.conf)
            self._changedetector.add_callback(self.restart)
        try:
            self._collect_services()
            for service in self._services.values():
                service.register_state_change_listener(
                    self._service_state_changed)
        except Exception as e:
            self.conf.log.exception(e)
            if self._changedetector:
                self._services.clear()
                if isinstance(e, SyntaxError):
                    self._changedetector.observe(e.filename)
                else:
                    for frame in traceback.extract_tb(e.__traceback__):
                        self._changedetector.observe(frame[0])
            raise

    def _service_state_changed(self, service, old, new):
        states = {service.name: service.state
                  for service in self._services.values()}
        self.trigger('state-change', states)
        if new == Service.State.EXCEPTION:
            self.conf.log.exception(service.exception)

    def _collect_services(self):
        self._services = OrderedDict()
        score = init_from_file(self.conf.conf)
        if self._changedetector:
            for file in parse_list(score.conf['score.init']['_files']):
                self._changedetector.observe(file)
        for desc in self.conf.modules:
            for name, worker in self._iter_workers(score, desc):
                self._services[name] = Service(name, worker)

    def _iter_workers(self, score, descriptor):
        if '/' in descriptor:
            module, names = tuple(map(str.strip, descriptor.split('/', 1)))
            names = tuple(map(str.strip, names.split(',')))
        else:
            module = descriptor
            names = None
        response = score._modules[module].score_serve_workers()
        if isinstance(response, Worker):
            yield module, response
        elif isinstance(response, list):
            if len(response) == 1:
                yield module, response[0]
                return
            for i, worker in enumerate(response):
                name = '%s/%d' % (module, i)
                yield name, worker
        elif isinstance(response, dict):
            for name, worker in response.items():
                if names is not None and name not in names:
                    continue
                name = '%s/%s' % (module, name)
                yield name, worker
        else:
            raise RuntimeError(
                'Invalid return value of %s.score_serve_workers(): %s' %
                (module, repr(response)))

    def restart(self, *_):
        self.trigger('restart')
        changedetector = self._changedetector
        self._changedetector = None
        if changedetector:
            changedetector.stop(wait=False)
        self.stop()

    def _call_on_subservices(self, func, *args):
        for service in self._services.values():
            getattr(service, func)(*args)



import abc
import threading

from .worker import Worker


class SimpleWorker(Worker):
    """
    A simplified worker base class, that hides all the ugly threading logic.

    You can subclass this and implement a ``loop`` function that periodically
    checks this.running:

    .. code-block:: python

        class Spammer(SimpleWorker):

            def loop():
                while self.running:
                    print('spam!')
                    time.sleep(1)
    """

    def prepare(self):
        self.__lock = threading.Lock()
        self.__thread = None

    def start(self):
        self.__thread = threading.Thread(target=self.__loop).start()

    def pause(self):
        with self.__lock:
            self.__running = False
        if self.__thread:
            self.__thread.join()
            self.__thread = None

    def stop(self):
        self.pause()

    def __loop(self):
        with self.__lock:
            self.__running = True
        try:
            self.loop()
        except Exception as e:
            self.service.set_exception(e)

    def cleanup(self, exception):
        pass

    @property
    def running(self):
        with self.__lock:
            return self.__running

    @abc.abstractmethod
    def loop(self):
        pass



import abc
import threading
import asyncio
from .worker import Worker
import concurrent.futures

try:
    from asyncio import run_coroutine_threadsafe
except ImportError:

    def run_coroutine_threadsafe(coro, loop):
        future = concurrent.futures.Future()

        def done(task_future):
            exception = task_future.exception()
            if exception:
                future.set_exception(exception)
            else:
                future.set_result(task_future.result())

        def queue_task():
            try:
                task_future = getattr(asyncio, 'async')(coro, loop=loop)
                task_future.add_done_callback(done)
            except Exception as exc:
                if future.set_running_or_notify_cancel():
                    future.set_exception(exc)
                raise

        loop.call_soon_threadsafe(queue_task)
        return future


class AsyncioWorker(Worker):
    """
    A specialized worker for :mod:`asyncio` servers.

    This base class will add a layer of abstraction to eliminate threading.
    Subclasses can override the functions :meth:`_prepare`, :meth:`_start`,
    :meth:`_pause`, :meth:`_stop` and :meth:`_cleanup`. These functions will be
    called inside a running event loop (which can be accessed as ``self.loop``)
    and can be regular functions or :term:`coroutines <coroutine>`.

    Example implementation:

    .. code-block:: python

        class EchoServer(AsyncioWorker):

            @asyncio.coroutine
            def _start(self):
                self.server = yield from self.loop.create_server(myserver)

            def _pause(self):
                self.server.close()
    """

    loop = None

    def prepare(self):
        if self.loop is None:
            self.loop = asyncio.new_event_loop()
        event = threading.Event()
        threading.Thread(target=self.__start_loop, args=(event,)).start()
        event.wait()
        event.clear()
        future = run_coroutine_threadsafe(self.__prepare(), self.loop)
        future.add_done_callback(lambda future: event.set())
        event.wait()
        exception = future.exception()
        if exception:
            raise exception

    def start(self):
        event = threading.Event()
        future = run_coroutine_threadsafe(self.__start(), self.loop)
        future.add_done_callback(lambda future: event.set())
        event.wait()
        exception = future.exception()
        if exception:
            raise exception

    def pause(self):
        event = threading.Event()
        future = run_coroutine_threadsafe(self.__pause(), self.loop)
        future.add_done_callback(lambda future: event.set())
        event.wait()
        exception = future.exception()
        if exception:
            raise exception

    def stop(self):

        def stop_loop(future):
            self.loop.call_soon_threadsafe(self.__stop_loop, event)

        event = threading.Event()
        future = run_coroutine_threadsafe(self.__stop(), self.loop)
        future.add_done_callback(stop_loop)
        event.wait()
        exception = future.exception()
        if exception:
            raise exception

    def cleanup(self, exception):
        if not self.loop.is_running():
            return

        def stop_loop(future):
            self.loop.call_soon_threadsafe(self.__stop_loop, event)

        event = threading.Event()
        future = run_coroutine_threadsafe(
            self.__cleanup(exception), self.loop)
        future.add_done_callback(stop_loop)
        event.wait()

    def _prepare(self):
        """
        Equivalent of :meth:`Worker.prepare`.

        This function will be called inside a running event loop.
        """
        pass

    @abc.abstractmethod
    def _start(self):
        """
        Equivalent of :meth:`Worker.start`.

        This function will be called inside a running event loop.
        """
        pass

    @abc.abstractmethod
    def _pause(self):
        """
        Equivalent of :meth:`Worker.pause`.

        This function will be called inside a running event loop.
        """
        pass

    def _stop(self):
        """
        Equivalent of :meth:`Worker.stop`.

        This function will be called inside a running event loop.
        """
        pass

    @abc.abstractmethod
    def _cleanup(self, exception):
        """
        Equivalent of :meth:`Worker.cleanup`.

        This function will be called inside a running event loop.
        """
        pass

    def __start_loop(self, event):
        event.set()
        self.loop.run_forever()

    @asyncio.coroutine
    def __prepare(self):
        result = self._prepare()
        if asyncio.iscoroutine(result):
            result = yield from result

    @asyncio.coroutine
    def __start(self):
        result = self._start()
        if asyncio.iscoroutine(result):
            result = yield from result

    @asyncio.coroutine
    def __pause(self):
        result = self._pause()
        if asyncio.iscoroutine(result):
            result = yield from result

    @asyncio.coroutine
    def __stop(self):
        result = self._stop()
        if asyncio.iscoroutine(result):
            result = yield from result

    @asyncio.coroutine
    def __cleanup(self, exception):
        result = self._cleanup(exception)
        if asyncio.iscoroutine(result):
            result = yield from result

    def __stop_loop(self, event):
        if not self.loop.is_running():
            event.set()
            return

        def stop(future=None):
            pending_tasks = [t for t in asyncio.Task.all_tasks(self.loop)
                             if not t.done()]
            if pending_tasks:
                task = pending_tasks.pop()
                task.add_done_callback(stop)
            else:
                self.loop.stop()
                event.set()

        stop()
